from . import engine
from concurrent.futures import ThreadPoolExecutor

def create(**options):
    model = engine.neuralNetworkModel(options.get("name"))
    model.layers = options.get("layers")
    model._init_weights()
    return model

def load(**options):
    model = engine.neuralNetworkModel(options.get("name"))
    model.load()
    return model

#Main backprop function with multithread support mini batch gradient descent
def mbgd(model, data_input, data_output, **options):

    #Get function parameters
    batch_size, epoch_number, learning_rate = options.get("batch_size"), options.get("epoch_number"), options.get("rate")

    #Function used to multithread each backprop of a batch inside the main backprop function
    def _backpropagate_batch(model, input, output, calculate_loss):
        run_history = model._run(input) #Run the model and save the history of each neuron output inside the network
        model._backpropagate(run_history, output, not(calculate_loss)) #Backprop over the network

    #Loop over each epoch
    for epoch in range(epoch_number):
        loss = 0 #initialize loss variable

        #Loop over the entire dataset, n-time where n is len(data)/batch_size, so an iteration for each batch
        for batch_start in range(0, len(data_input), batch_size): #batch_start is the index of the start of the batch
            batch_end =  min(batch_start + batch_size,len(data_input)) #batch_end is the index of the end of the batch
            model._init_batch() #Call init_batch() before running the backprop, initializes the space to store weights and biases modifications

            #Multithread each single run and backpropagation of a batch
            with ThreadPoolExecutor(max_workers=batch_size) as executor:
                for i in range(batch_start, batch_end): #Loop over each single input inside a batch
                    executor.submit(_backpropagate_batch, model, data_input[i], data_output[i], i-batch_start) #Call _backpropagate_batch() for each thread

            #extract the average loss of the batch (simulated average for speed optimization)
            loss += model.loss
            #Update weights and biases of the model based on the backpropagation
            model._update_weights(batch_size, learning_rate)

        #Print median loss for each epoch
        print("Epoch = " + str(epoch) + " Loss = " + str(loss/(len(data_input)/batch_size)), end='\r')

    #Save the model
    model.save()
    print("\nTraining finished, model saved.")



import numpy as np
import threading

class neuralNetworkModel:
    def __init__(self, name = "unknown"):
        self.name = name

    def _run(self, input):
        layer_history = np.empty(len(self.layers), dtype=object)
        layer_history[0] = input
        for layer in range(1,len(self.layers)):
            layer_history[layer] = np.add(np.dot(self.weights[layer], layer_history[layer-1]), self.biases[layer])
            match self.layers[layer][1]:
                case "sigmoid": layer_history[layer] = np.float32(1) / (np.float32(1) + np.exp(-layer_history[layer]))
                case "relu": layer_history[layer] = np.maximum(np.float32(0), layer_history[layer])
                case "leakyRelu": layer_history[layer] = np.maximum(np.float32(0.1) * layer_history[layer], layer_history[layer])
        return layer_history

    def _backpropagate(self, run_history, output, calculate_loss):
        layer_number = len(self.layers)
        error = np.subtract(run_history[-1], output)
        for layer in range(layer_number - 1, 0, -1):
            match self.layers[layer][1]:
                case "sigmoid": delta = np.multiply(error, run_history[layer] * (np.float32(1) - run_history[layer]))
                case "relu": delta = np.multiply(error, np.where(run_history[layer] > np.float32(0), np.float32(1), np.float32(0)))
                case "leakyRelu": delta = np.multiply(error, np.where(run_history[layer] > np.float32(0), np.float32(1), np.float32(0.1)))
            with threading.Lock():
                self.weights_mod[layer] += np.outer(delta, run_history[layer - 1])
                self.biases_mod[layer] += delta
            if layer>1: error = np.dot(self.weights[layer].T, delta)
        if calculate_loss:
            self.loss = abs(np.mean(run_history[-1] - output))

    def _update_weights(self, batch_size, learning_rate):
        factor = np.float32(1/batch_size*learning_rate)
        self.weights -= self.weights_mod * factor
        self.biases -= self.biases_mod * factor

    def _init_weights(self):
        self.weights, self.biases = np.zeros(len(self.layers), dtype=object), np.zeros(len(self.layers), dtype=object)
        for i in range(1, len(self.layers)):
            match self.layers[i][1]:
                case "sigmoid": self.weights[i] = np.random.uniform(low=-0.1, high=0.1, size=(self.layers[i][0], self.layers[i-1][0])).astype(np.float32)
                case "relu": self.weights[i] = np.random.normal(loc=0, scale=np.sqrt(2 / self.layers[i-1][0]), size=(self.layers[i][0], self.layers[i-1][0])).astype(np.float32)
                case "leakyRelu": self.weights[i] = np.random.normal(loc=0, scale=np.sqrt(2 / self.layers[i-1][0]), size=(self.layers[i][0], self.layers[i-1][0])).astype(np.float32)
            self.biases[i] = np.zeros(self.layers[i][0], dtype=np.float32)

    def _init_batch(self):
        self.weights_mod, self.biases_mod, self.loss = np.zeros_like(self.weights), np.zeros_like(self.biases), 0

    def run(self, input):
        return self._run(input)[-1]

    def save(self):
        np.savez(self.name + ".npz", matrix1=self.weights, matrix2=self.biases, matrix3=self.layers)

    def load(self):
        data = np.load(self.name + ".npz", allow_pickle=True)
        self.weights, self.biases, self.layers = data['matrix1'], data['matrix2'], [[int(item) if item.isdigit() else item for item in row] for row in data['matrix3']]

    def pop(self, index=-1):
        self.layers.pop(index)
        self.weights = np.delete(self.weights, index, axis=0)
        self.biases = np.delete(self.biases, index, axis=0)

    def insert(self, model, index=-1):
        self.layers[index:index] = model.layers
        self.weights = np.insert(self.weights, index, model.weights, axis=0)
        self.biases = np.insert(self.biases, index, model.biases, axis=0)



import cv2
import numpy as np
import os
import gzip

def saveImage(vector, image_path, image_size, color):
    if color == "grayscale":
        vector = (vector * 255.0).reshape(tuple(reversed(image_size))).astype(np.uint8)
        vector = cv2.cvtColor(vector, cv2.COLOR_GRAY2BGR)
    else:
        vector = (vector * 255.0).reshape(image_size + (3,)).astype(np.uint8)
    cv2.imwrite(image_path, vector)

def loadImage(image_path, desired_size, color):
    image = cv2.imread(image_path)
    image = cv2.resize(image, desired_size)
    if color == "grayscale":
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    vectorized_image = image.flatten() / 255.0
    return vectorized_image.astype(np.float32)

def decimal_to_array(decimal_number, n):
    binary_array = np.zeros(n, dtype=np.float32)
    binary_array[decimal_number] = 1
    return binary_array

def load(**options):
    data_type = options.get("data_type")
    match data_type: 
        case "image":
            folder = options.get("folder")
            resize = options.get("resize")
            file_names = os.listdir(folder)
            all_images_vector = []
            for file_name in file_names:
                image_vector = loadImage(folder + "/" + file_name, resize, options.get("color"))
                all_images_vector.append(image_vector)
            return np.vstack(all_images_vector)
        case "gz":
            path = options.get("path")
            with gzip.open(path, 'rb') as f:
                file_content = f.read()
            data = np.frombuffer(file_content, dtype=np.uint8)[options.get("start_index"):]
            if options.get("input_number") != None:
                data = data.reshape((-1, options.get("input_number")))
            if options.get("divide") != None:
                data = data/options.get("divide")
            if options.get("one_hot") != None:
                Q =  []
                for i in range(len(data)):
                    Q.append(decimal_to_array(data[i], int(options.get("one_hot"))))
                data = np.vstack(Q)
            return data.astype(np.float32)
        case "numbers":
            path = options.get("path")
            with open(path, 'r', encoding='utf-8') as file:
                n_arr = file.readlines()
                n_arr = [float(x) for x in n_arr]
            data = np.array(n_arr)
            return data



import os
from setuptools import setup

directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(directory, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

setup(
    name='pileoffeather',
    version='0.4.2',
    license='MIT',
    url = 'https://github.com/usedToBeTomas/pile-of-feather',
    author='Daniele Tomaselli',
    description='Lightweight and easy to use ml library for small projects, create a neural network in minutes.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['pileoffeather'],
    python_requires='>=3.6',
    keywords = ['neural network', 'ml', 'ai', 'machine learning','simple','nn'],
    install_requires=[
        'opencv-python',
        'numpy'
    ]
)
